import pytest

from nitrate.passwords import get_required_password, use_in_memory_keyring


class TestGetRequiredPassword:
    def test_gets_existing_password(self) -> None:
        use_in_memory_keyring(initial_passwords={("flickr", "api_key"): "12345"})

        assert get_required_password("flickr", "api_key") == "12345"

    def test_throws_if_password_does_not_exist(self) -> None:
        with pytest.raises(RuntimeError, match="Could not retrieve password"):
            get_required_password("doesnotexist", "doesnotexist")



"""
Some utilities for working with XML in a type-safe way.
"""

import xml.etree.ElementTree as ET


def find_required_elem(elem: ET.Element, *, path: str) -> ET.Element:
    """
    Find the first subelement matching ``path``, or throw if absent.
    """
    # We use this when we're parsing responses from the Flickr API, and
    # there are certain elements we know should always be present in responses.
    #
    # e.g. we know that photos always have a <title>, so we could write:
    #
    #     photo_elem.find(".//title")
    #
    # But the type checker only knows that ``find()`` returns Optional[Element] --
    # it doesn't know that this path should always be present in the response.
    #
    # If we call it from this function instead:
    #
    #     find_required(photo_elem, path=".//title")
    #
    # Then the type checker can see that it returns a well-defined Element,
    # and it's happy for us to use it without checking in the rest of the code.
    matching_elem = elem.find(path=path)

    if matching_elem is None:
        raise ValueError(f"Could not find required match for {path!r} in {elem!r}")

    return matching_elem


def find_required_text(elem: ET.Element, *, path: str) -> str:
    """
    Find the text of the first element matching ``path``, or throw if absent.
    """
    # We use this when we're parsing responses from the Flickr API, and
    # there are certain elements we know should always be present and have text.
    #
    # e.g. we know that users always have a <id> with some text, so we could write:
    #
    #     user_elem.find(".//id")
    #
    # But the type checker only knows that ``find()`` returns Optional[Element] --
    # it doesn't know that this path should always be present in the response.
    #
    # If we call it from this function instead:
    #
    #     find_required_text(user_elem, path=".//id")
    #
    # Then the type checker can see that it returns a well-defined Element,
    # and it's happy for us to use it without checking in the rest of the code.
    matching_elem = find_required_elem(elem, path=path)
    text = matching_elem.text

    if text is None:
        raise ValueError(f"Could not find required text in {matching_elem}")

    return text


def find_optional_text(elem: ET.Element, *, path: str) -> str | None:
    """
    Find the text of the first element matching ``path``, or return None
    if the element is missing, has no text, or has empty text.
    """
    matching_elem = elem.find(path=path)

    if matching_elem is None:
        return None

    return matching_elem.text or None


__all__ = ["find_required_elem", "find_required_text", "find_optional_text"]



import datetime
import pathlib
import typing

import pytest
from pydantic import ValidationError

from nitrate.json import DatetimeDecoder
from nitrate.types import read_typed_json, validate_type


class Shape(typing.TypedDict):
    color: str
    sides: int


@pytest.mark.parametrize(
    "data",
    [
        {"color": "red"},
        {"sides": 4},
        {"color": "red", "sides": "four"},
        {"color": (255, 0, 0), "sides": 4},
        {"color": "red", "sides": 4, "angle": 36},
    ],
)
def test_validate_type_flags_incorrect_data(data: typing.Any) -> None:
    with pytest.raises(ValidationError):
        validate_type(data, model=Shape)


def test_validate_type_allows_valid_data() -> None:
    validate_type({"color": "red", "sides": 4}, model=Shape)


def test_validate_type_supports_builtin_types() -> None:
    validate_type([1, 2, 3], model=list[int])


def test_read_typed_json_allows_valid_data(tmp_path: pathlib.Path) -> None:
    json_path = tmp_path / "data.json"

    with open(json_path, "w") as out_file:
        out_file.write("[1, 2, 3]")

    assert read_typed_json(json_path, model=list[int]) == [1, 2, 3]


def test_read_typed_json_flags_invalid_data(tmp_path: pathlib.Path) -> None:
    json_path = tmp_path / "data.json"

    with open(json_path, "w") as out_file:
        out_file.write("[1, 2, 3]")

    with pytest.raises(ValidationError):
        read_typed_json(json_path, model=dict[str, str])


def test_read_typed_json_uses_decoder(tmp_path: pathlib.Path) -> None:
    json_path = tmp_path / "data.json"

    with open(json_path, "w") as out_file:
        out_file.write(
            '{"date": {"type": "datetime.datetime", "value": "2023-12-27T14:16:02Z"}}'
        )

    with pytest.raises(ValidationError):
        read_typed_json(json_path, model=dict[str, datetime.datetime])

    expected = {
        "date": datetime.datetime(2023, 12, 27, 14, 16, 2, tzinfo=datetime.timezone.utc)
    }

    actual = read_typed_json(
        json_path, model=dict[str, datetime.datetime], cls=DatetimeDecoder
    )

    assert actual == expected



"""
Allow encoding/decoding JSON in a type-preserving way.

In particular, these classes allow us to store ``datetime.datetime``
values in JSON and retrieve them as proper datetime objects, rather
than e.g. strings.
"""

import datetime
import json
import typing


T = typing.TypeVar("T")

EncodedDate = typing.TypedDict(
    "EncodedDate", {"type": typing.Literal["datetime.datetime"], "value": str}
)


class DatetimeEncoder(json.JSONEncoder):
    """
    A custom JSON encoder that supports datetimes.

        >>> t = datetime.datetime(2001, 2, 3, 4, 5, 6)
        >>> json.dumps({"t": t}, cls=DatetimeEncoder)
        '{"t": {"type": "datetime.datetime", "value": "2001-02-03T04:05:06"}}'

    This is meant to be used with ``DatetimeDecoder`` -- together, they
    allow you to serialise a datetime value via JSON and preserve its type.

    """

    def default(self, t: T) -> typing.Any:
        if isinstance(t, datetime.datetime):
            return {"type": "datetime.datetime", "value": t.isoformat()}
        else:
            return super().default(t)


class DatetimeDecoder(json.JSONDecoder):
    """
    A custom JSON decoder that supports the datetimes encoded
    by DatetimeEncoder.

        >>> json.loads(
        ...     '{"t": {"type": "datetime.datetime", "value": "2001-02-03T04:05:06"}}',
        ...     cls=DatetimeDecoder)
        {'t': datetime.datetime(2001, 2, 3, 4, 5, 6)}

    """

    def __init__(self) -> None:
        super().__init__(object_hook=self.dict_to_object)

    def dict_to_object(
        self, d: dict[str, typing.Any]
    ) -> dict[str, typing.Any] | datetime.datetime:
        if d.get("type") == "datetime.datetime":
            return datetime.datetime.fromisoformat(d["value"])
        else:
            return d


__all__ = ["DatetimeEncoder", "DatetimeDecoder"]



import httpx
import pytest
from vcr.cassette import Cassette

from nitrate.cassettes import get_cassette_name


def test_creates_cassette(cassette_name: str) -> None:
    assert cassette_name == "test_creates_cassette.yml"


@pytest.mark.parametrize(
    ["expected_cassette_name"],
    [
        pytest.param("test_creates_parametrized_cassette[test1].yml", id="test1"),
        pytest.param("test_creates_parametrized_cassette[test2].yml", id="test2"),
        pytest.param("test_creates_parametrized_cassette[test3].yml", id="test3"),
    ],
)
def test_creates_parametrized_cassette(
    cassette_name: str, expected_cassette_name: str
) -> None:
    assert cassette_name == expected_cassette_name


class TestCassetteNameInClass:
    def test_prefixes_class_name_to_cassette(self, cassette_name: str) -> None:
        assert (
            cassette_name
            == "TestCassetteNameInClass.test_prefixes_class_name_to_cassette.yml"
        )

    @pytest.mark.parametrize(
        ["expected_cassette_name"],
        [
            pytest.param(
                "TestCassetteNameInClass.test_prefixes_name_with_parametrized_cassette[test1].yml",
                id="test1",
            ),
            pytest.param(
                "TestCassetteNameInClass.test_prefixes_name_with_parametrized_cassette[test2].yml",
                id="test2",
            ),
            pytest.param(
                "TestCassetteNameInClass.test_prefixes_name_with_parametrized_cassette[test3].yml",
                id="test3",
            ),
            pytest.param(
                "TestCassetteNameInClass.test_prefixes_name_with_parametrized_cassette[test.name.with.periods].yml",
                id="test.name.with.periods",
            ),
        ],
    )
    def test_prefixes_name_with_parametrized_cassette(
        self, cassette_name: str, expected_cassette_name: str
    ) -> None:
        assert cassette_name == expected_cassette_name


@pytest.mark.parametrize("url", ["https://example.com"])
def test_throws_if_bad_cassette_name(url: str, request: pytest.FixtureRequest) -> None:
    with pytest.raises(ValueError, match="Illegal characters in VCR cassette name"):
        get_cassette_name(request)


def test_creates_cassette_in_fixture_dir(vcr_cassette: Cassette) -> None:
    resp = httpx.get("https://example.com")
    resp.raise_for_status()

    assert (
        vcr_cassette._path
        == "tests/fixtures/cassettes/test_creates_cassette_in_fixture_dir.yml"
    )



"""
Use cassettes for testing HTTP APIs with VCR.py [1]

This establishes a couple of conventions for where cassettes are stored
and how they're named.

[1]: https://vcrpy.readthedocs.io/
"""

from collections.abc import Iterator

import pytest
import vcr
from vcr.cassette import Cassette


def get_cassette_name(request: pytest.FixtureRequest) -> str:
    """
    Returns the name of a cassette for vcr.py.

    The name can be made up of (up to) three parts:

    -   the name of the test class
    -   the name of the test function
    -   the ID of the test case in @pytest.mark.parametrize

    """
    name = request.node.name

    # This is to catch cases where e.g. we try to include a complete
    # HTTP URL in a cassette name, which creates very messy folders in
    # the fixtures directory.
    if any(char in name for char in ":/"):
        raise ValueError(
            "Illegal characters in VCR cassette name - "
            "please set a test ID with pytest.param(…, id='…')"
        )

    if request.cls is not None:
        return f"{request.cls.__name__}.{name}.yml"
    else:
        return f"{name}.yml"


@pytest.fixture(scope="function")
def cassette_name(request: pytest.FixtureRequest) -> str:
    return get_cassette_name(request)


@pytest.fixture(scope="function")
def vcr_cassette(cassette_name: str) -> Iterator[Cassette]:
    """
    Creates a VCR cassette for use in tests.

    Anything using httpx in this test will record its HTTP interactions
    as "cassettes" using vcr.py, which can be replayed offline
    (e.g. in CI tests).
    """
    with vcr.use_cassette(
        cassette_name,
        cassette_library_dir="tests/fixtures/cassettes",
        decode_compressed_response=True,
    ) as cassette:
        yield cassette


__all__ = ["cassette_name", "vcr_cassette"]



import datetime
import json
import uuid

import pytest

from nitrate.json import DatetimeDecoder, DatetimeEncoder


def test_can_json_round_trip() -> None:
    d = {
        "label": "an interesting time",
        "time": datetime.datetime(2001, 2, 3, 4, 5, 6),
    }

    json_string = json.dumps(d, cls=DatetimeEncoder)
    parsed_json_string = json.loads(json_string, cls=DatetimeDecoder)

    assert parsed_json_string == d


def test_an_unrecognised_type_still_fails() -> None:
    with pytest.raises(TypeError, match="Object of type UUID is not JSON serializable"):
        json.dumps({"id": uuid.uuid4()})

    with pytest.raises(TypeError, match="Object of type UUID is not JSON serializable"):
        json.dumps({"id": uuid.uuid4()}, cls=DatetimeEncoder)
